@@ -1,7 +1,8 @@
 package Row;
 use ObjStore;
-use base 'ObjStore::AVHV';
-use Class::Fields qw(f1 f2 f3);
+use base 'ObjStore::HV';
+
+sub _is_corrupted {0}
 
 # Gee, that was hard.
 
